#pragma once

#define ARON_MAJOR "1"
#define ARON_MINOR "1"
#define ARON_PATCH "0"
#define ARON_VERSION "1.1.0" // allow templates + any types

module armarx
{
    module aron
    {

        const string VERSION = ARON_VERSION;

        /*************************
         * General Definitions ***
         ************************/
        sequence<byte> AronByteSeq;
        sequence<int> AronIntSeq;
        sequence<string> AronStringSeq;
        dictionary<string, int> AronIntDict;


        /*************************
         * Aron Types ************
         ************************/
        module type
        {
            enum Maybe
            {
                eNone,
                eOptional,
                eRawPointer,
                eUniquePointer,
                eSharedPointer
            };

            module ndarray
            {
                enum ElementType
                {
                    uint8,
                    int8,
                    uint16,
                    int16,
                    uint32,
                    int32,
                    float32,
                    float64
                };
            };

            module image
            {
                enum PixelType
                {
                    rgb24,
                    depth32
                };
            };

            module pointcloud
            {
                enum VoxelType
                {
                    PointXYZ,
                    PointXYZI,
                    PointXYZL,
                    PointXYZRGB,
                    PointXYZRGBL,
                    PointXYZRGBA,
                    PointXYZHSV
                };
            };

            module matrix
            {
                enum ElementType
                {
                    int16,
                    int32,
                    int64,
                    float32,
                    float64
                };
            };

            module quaternion
            {
                enum ElementType
                {
                    float32,
                    float64
                };
            };

            module dto
            {
                class GenericType {
                    string VERSION = ARON_VERSION;
                    Maybe maybe = Maybe::eNone;
                }
                sequence<GenericType> GenericTypeSeq;
                dictionary<string, GenericType> GenericTypeDict;

                /* ***** Container types ***** */
                class List extends GenericType { GenericType acceptedType; }
                class Tuple extends GenericType { GenericTypeSeq elementTypes; }
                class Pair extends GenericType { GenericType acceptedType1; GenericType acceptedType2; }
                class AronObject extends GenericType { AronObject parent; AronStringSeq templates; string objectName; AronStringSeq templateInstantiations; GenericTypeDict elementTypes; }
                class Dict extends GenericType { GenericType acceptedType; }

                /* ***** Complex Types (serialize to ndarray) ***** */
                class NDArray extends GenericType { int ndimensions; ndarray::ElementType elementType; }
                class Matrix extends GenericType { int rows; int cols; matrix::ElementType elementType; }
                class Quaternion extends GenericType { quaternion::ElementType elementType; }
                class Image extends GenericType { image::PixelType pixelType; }
                class PointCloud extends GenericType { pointcloud::VoxelType voxelType; }
                class Position extends GenericType { }
                class Orientation extends GenericType { }
                class Pose extends GenericType { }

                /* ***** Enum types ***** */
                class IntEnum extends GenericType { string enumName; AronIntDict acceptedValues; }

                /* ***** Primitive Types ***** */
                class AronInt extends GenericType { };
                class AronLong extends GenericType { };
                class AronDouble extends GenericType { };
                class AronFloat extends GenericType { };
                class AronString extends GenericType { };
                class AronBool extends GenericType { };
                class AronTime extends GenericType { };

                /* ***** Any Types ***** */
                class AnyObject extends GenericType {};
            };
        };


        /*************************
         * Aron Data *************
         ************************/
        module data
        {
            module dto
            {
                class GenericData {
                    string VERSION = ARON_VERSION;
                };
                sequence<GenericData> GenericDataSeq;
                dictionary<string, GenericData> GenericDataDict;

                /* ***** Container Data ***** */
                class List extends GenericData { GenericDataSeq elements; };
                class Dict extends GenericData { GenericDataDict elements; };

                /* ***** Complex Data ***** */
                // The NDArray contains more or less the same information as an AronType, but there is no other way to do it
                // Especially, note the difference between the type's typeName (e.g. "GRAY_SCALE_IMAGE" => language dependent) and the data's type ("0")
                // Further, note the difference between the type's dimensions (e.g. 128x128) and the data's dimensions (128x128x3 for RGB)
                class NDArray extends GenericData { AronIntSeq shape; string type; AronByteSeq data; }


                /* ***** Primitive Data ***** */
                class AronInt extends GenericData { int value; };
                class AronLong extends GenericData { long value; };
                class AronDouble extends GenericData { double value; };
                class AronFloat extends GenericData { float value; };
                class AronString extends GenericData { string value; };
                class AronBool extends GenericData { bool value; };

                // useful for memory ice_conversions
                sequence<Dict> AronDictSeq;
            };
        };
    };
};
